import logging
from types import TracebackType
from typing import Any, Dict, Optional, Self, Type, Union

import httpx
from httpx import HTTPError

from snakestack.httpx.exceptions import RequestHTTPError

logger = logging.getLogger(__name__)


class SnakeHttpClient:

    def __init__(self: Self, base_url: str) -> None:
        self.base_url = base_url
        self.client = httpx.AsyncClient(base_url=base_url)

    async def handle(
        self: Self,
        method: str,
        url: str,
        content: Optional[Union[str, bytes]] = None,
        data: Optional[Dict[str, Any]] = None,
        files: Optional[Dict[str, Any]] = None,
        json: Optional[Any] = None,
        params: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, str]] = None,
        cookies: Optional[Dict[str, str]] = None,
        auth: Optional[Any] = None,
        follow_redirects: bool = False,
        timeout: Optional[Union[float, httpx.Timeout]] = None,
        extensions: Optional[Dict[str, Any]] = None,
    ) -> httpx.Response:
        try:
            response = await self.client.request(
                method=method,
                url=url,
                content=content,
                data=data,
                files=files,
                json=json,
                params=params,
                headers=headers,
                cookies=cookies,
                auth=auth,
                follow_redirects=follow_redirects,
                timeout=timeout,
                extensions=extensions
            )
            response.raise_for_status()
        except HTTPError as error:
            logger.exception("Error on request", exc_info=error)
            raise RequestHTTPError(api=f"{self.base_url}{url}", original_exception=error)
        else:
            return response

    async def aclose(self: Self) -> None:
        await self.client.aclose()

    async def __aenter__(self: Self) -> Self:
        return self

    async def __aexit__(
        self: Self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        await self.aclose()



from contextvars import ContextVar

request_id_var: ContextVar[str | None] = ContextVar("request_id", default=None)

def set_request_id(request_id: str) -> None:
    request_id_var.set(request_id)

def get_request_id() -> str | None:
    return request_id_var.get()

def reset_request_id() -> None:
    request_id_var.set(None)



from math import ceil
from typing import Generic, Self, TypeVar

from pydantic import BaseModel, ConfigDict, Field, model_validator

from snakestack import constants

T = TypeVar("T")


class PaginatedModel(BaseModel, Generic[T]):
    model_config = ConfigDict(
        extra='forbid'
    )

    items: list[T] = Field(
        ..., description=constants.PAGINATE_ITEMS_SCHEMA_DESCRIPTION
    )
    total: int = Field(
        ..., description=constants.PAGINATE_TOTAL_SCHEMA_DESCRIPTION
    )
    page: int = Field(
        default=1, description=constants.PAGINATE_PAGE_SCHEMA_DESCRIPTION
    )
    size: int = Field(
        ..., description=constants.PAGINATE_SIZE_SCHEMA_DESCRIPTION
    )
    has_next: bool = Field(
        default=False, description=constants.PAGINATE_HAS_NEXT_SCHEMA_DESCRIPTION
    )
    has_prev: bool = Field(
        default=False, description=constants.PAGINATE_HAS_PREV_SCHEMA_DESCRIPTION
    )
    total_pages: int = Field(
        default=0,
        description=constants.PAGINATE_TOTAL_PAGES_SCHEMA_DESCRIPTION
    )
    offset: int = Field(
        default=0,
        description=constants.PAGINATE_OFFSET_SCHEMA_DESCRIPTION
    )

    @model_validator(mode="after")
    def validate_page(self: Self) -> Self:
        self.page = (self.offset // self.size) + 1
        return self

    @model_validator(mode="after")
    def calculate_pagination_metadata(
        self: Self
    ) -> Self:
        self.total_pages = ceil(self.total / self.size)
        self.has_next = self.page < self.total_pages
        self.has_prev = self.page > 1
        return self



from .helpers import (
    get_current_span,
    get_tracer,
    instrumented_span,
    set_span_attributes,
)
from .instrumentor import instrument_app

__all__ = ["instrument_app", "get_tracer", "instrumented_span", "get_current_span", "set_span_attributes"]



import asyncio
import logging
from asyncio.events import AbstractEventLoop
from typing import Self, cast

try:
    from google.cloud.pubsub_v1 import SubscriberClient
    from google.cloud.pubsub_v1.subscriber.futures import StreamingPullFuture
    from google.cloud.pubsub_v1.types import FlowControl, SubscriberOptions
except ImportError:
    raise RuntimeError("Pubsub extra is not installed. Run `pip install snakestack[pubsub]`.")

from .batch import BatchesType, BatchingWorker, BatchSettingsModel
from .exceptions import AckableError, RetryableError
from .processors import BatchProcessor, SimpleProcessor
from .types import (
    AsyncQueueMessagePubSub,
    MessagePubSubType,
    QueuesType,
    SchemaType,
)
from .utils import get_message_info, parse_message

logger = logging.getLogger(__name__)

class SnakeStackSubscriber:

    def __init__(
        self: Self,
        client: "SubscriberClient",
        processor: SimpleProcessor | BatchProcessor,
        subscription_path: str,
        flow_control: FlowControl,
        schema: SchemaType | None = None
    ) -> None:
        self.client: "SubscriberClient" = client
        self.subscription_path: str = subscription_path
        self._max_latency_ms: float | None = None
        self._streaming_pull_future: StreamingPullFuture | None = None
        self._schema: SchemaType | None = schema
        self._contexts: list[BatchSettingsModel] = []
        self._queue: AsyncQueueMessagePubSub
        self._batches: BatchesType = {}
        self._queues: QueuesType = {}
        self.processor = processor
        self._flow_control: FlowControl = flow_control
        self._is_batch: bool = hasattr(processor, "process_batch")
        self._stop_event: asyncio.Event = asyncio.Event()
        self._loop: AbstractEventLoop | None = None

    @classmethod
    def from_processor(
        cls: type[Self],
        processor: SimpleProcessor | BatchProcessor,
        project_id: str,
        subscription_name: str,
        *,
        flow_control: FlowControl,
        schema: SchemaType | None = None,
        enable_open_telemetry: bool = False
    ) -> Self:
        subscriber_options = SubscriberOptions(
            enable_open_telemetry_tracing=enable_open_telemetry
        )
        client = SubscriberClient(subscriber_options=subscriber_options)
        subscription_path = client.subscription_path(project_id, subscription_name)
        return cls(
            client=client,
            processor=processor,
            subscription_path=subscription_path,
            schema=schema,
            flow_control=flow_control,
        )

    async def start(
        self: Self,
        max_latency_ms: float | None = None,
        num_workers: int = 2,
        contexts: list[BatchSettingsModel] | None = None
    ) -> None:
        self._max_latency_ms = max_latency_ms
        self._contexts = contexts or []
        self._batches = {}
        self._loop = asyncio.get_running_loop()

        logger.debug(f"Starting consumer with {self.subscription_path}")

        async def _start() -> None:
            logger.debug("Starting streaming pull")
            self._queue = asyncio.Queue(maxsize=self._flow_control.max_messages or 100)

            self._streaming_pull_future = self.client.subscribe(
                subscription=self.subscription_path,
                callback=self._wrapper,
                flow_control=self._flow_control
            )
            logger.debug("Streaming pull started")

            for i in range(num_workers):
                logger.debug(f"Scheduling worker [{i}]")
                asyncio.create_task(self.worker(num=i))

            if isinstance(self.processor, BatchProcessor):
                for context in self._contexts:
                    batch = BatchingWorker(
                        batch_settings=context,
                        callback=self.processor.process_batch
                    )
                    queue: AsyncQueueMessagePubSub = asyncio.Queue(maxsize=context.size * 2)
                    asyncio.create_task(batch.run(queue=queue))
                    self._batches[context.name] = batch
                    self._queues[context.name] = queue

                self.processor.set_queues(self._queues)

        await _start()
        await self._stop_event.wait()
        await self.stop(streaming_pull_future=self._streaming_pull_future)
        logger.debug("Shutting down...")

    def shutdown(self: Self) -> None:
        logger.debug("Shutdown signal received.")
        self._stop_event.set()

    async def stop(self: Self, streaming_pull_future: StreamingPullFuture) -> None:
        logger.debug(f"Finishing consumer with {self.subscription_path}")
        if streaming_pull_future:
            streaming_pull_future.cancel()

        for batch in self._batches.values():
            await batch._flush()

    def _wrapper(self: Self, message: MessagePubSubType) -> None:
        logger.debug(f"Received raw message {message.message_id}")

        def enqueue() -> None:
            logger.debug(f"Queue size before enqueue: {self._queue.qsize()}")
            try:
                queue = self._queue
                queue.put_nowait(message)
            except asyncio.QueueFull:
                logger.warning(f"Queue is full. Nacking message {message.message_id}")
                message.nack()

        cast(AbstractEventLoop, self._loop).call_soon_threadsafe(enqueue)

    async def worker(self: Self, num: int) -> None:
        logger.debug(f"Starting worker [{num}]")
        while True:
            logger.debug(f"Worker [{num}] waiting messages...")
            queue = self._queue
            message = await queue.get()
            logger.debug(f"Worker [{num}] pulled message {message.message_id}")
            await self.process_message(message)

    async def process_message(self: Self, message: MessagePubSubType) -> None:
        try:
            _, _ = get_message_info(message=message)
            parsed_message = parse_message(message=message, schema=self._schema)
            await self.processor.process(data=parsed_message, message=message)
            if not self._is_batch:
                message.ack()
        except AckableError as error:
            logger.exception("Ackable error - Message will be discarded", exc_info=error)
            message.ack()
        except RetryableError as error:
            logger.exception("Retryable error - Message will be enqueued", exc_info=error)
            message.nack()
        except Exception as error:
            logger.exception("Unexpected error - Message will be discarded", exc_info=error)
            message.ack()



import logging
from typing import Any, Callable

try:
    from motor.motor_asyncio import AsyncIOMotorCollection
except ImportError:
    raise RuntimeError("Mongodb extra is not installed. Run `pip install snakestack[mongodb]`.")

from snakestack.mongodb.tracing import traced_motor_method

logger = logging.getLogger(__name__)


def patch_motor_collection_methods() -> None:
    methods_to_patch = [
        "find",
        "find_one",
        "insert_one",
        "insert_many",
        "update_one",
        "update_many",
        "delete_one",
        "delete_many",
        "replace_one",
        "aggregate",
    ]

    for method_name in methods_to_patch:
        if hasattr(AsyncIOMotorCollection, method_name):
            original = getattr(AsyncIOMotorCollection, method_name)
            patched = traced_motor_method(method_name)(original)
            if not is_already_patched(original):
                logger.debug("MongoDB collection method '%s' patched for tracing.", method_name)
                setattr(AsyncIOMotorCollection, method_name, patched)


def is_already_patched(method: Callable[..., Any]) -> bool:
    return getattr(method, "_is_traced_motor_method", False)



from typing import Any, cast

try:
    from redis.asyncio import Redis
except ImportError:
    raise RuntimeError("Redis extra is not installed. Run `pip install snakestack[redis]`.")

def create_async_redis_client(
    host: str = "localhost",
    port: int = 6379,
    db: int = 0,
    username: str | None = None,
    password: str | None = None,
    decode_responses: bool = True,
    **kwargs: Any,
) -> Redis:
    redis = Redis(
        host=host,
        port=port,
        db=db,
        username=username,
        password=password,
        decode_responses=decode_responses,
        **kwargs,
    )
    return cast(Redis, redis)



# Unidades de tamanho
KILOBYTE = 1024
MEGABYTE = 1024 * 1024
GIGABYTE = 1024 * 1024 * 1024
TERABYTE = 1024 * 1024 * 1024 * 1024

# Unidades de tempo
MILLISECONDS = 0.001
SECONDS = 1
MINUTES = 60 * SECONDS
HOURS = 60 * MINUTES

# Docs
ERROR_DETAIL_SCHEMA_DESCRIPTION = "Error message or explanation"
PAGINATE_LIMIT_SCHEMA_DESCRIPTION = "The maximum number of items to return per page."
PAGINATE_OFFSET_SCHEMA_DESCRIPTION = (
    "The number of items to skip before starting "
    "to collect the result set."
)
PAGINATE_TOTAL_PAGES_SCHEMA_DESCRIPTION = (
    "Total number of pages based on total items and page size."
)
PAGINATE_HAS_PREV_SCHEMA_DESCRIPTION = (
    "Indicates if there is a previous page available."
)
PAGINATE_HAS_NEXT_SCHEMA_DESCRIPTION = "Indicates if there is a next page available."
PAGINATE_SIZE_SCHEMA_DESCRIPTION = (
    "Total number of pages based on total items and page size."
)
PAGINATE_PAGE_SCHEMA_DESCRIPTION = "Current page number, starting at 1."
PAGINATE_TOTAL_SCHEMA_DESCRIPTION = "Total number of data available in the database."
PAGINATE_ITEMS_SCHEMA_DESCRIPTION = "List of resources returned in the current page."

SCHEMA_CHECK_OK_DESCRIPTION = (
    "Indicates whether the individual "
    "check passed successfully."
)
SCHEMA_CHECK_LATENCY_MS_DESCRIPTION = (
    "Latency in milliseconds for the check to complete."
)
SCHEMA_CHECK_ERROR_DESCRIPTION = (
    "Error message if the check failed; null if successful."
)
SCHEMA_HEALTHZ_SERVICE_NAME_DESCRIPTION = (
    "Name of the service responding to the ping request."
)
SCHEMA_HEALTHZ_VERSION_DESCRIPTION = (
    "Version of the deployed service or application."
)
SCHEMA_HEALTHZ_HOST_DESCRIPTION = (
    "Hostname or identifier of the running instance."
)
SCHEMA_HEALTHZ_UPTIME_DESCRIPTION = (
    "Formatted uptime of the service (e.g., '2d 4h')."
)
SCHEMA_HEALTHZ_TIMESTAMP_DESCRIPTION = "Current timestamp in ISO format."
SCHEMA_HEALTHZ_ENVIRONMENT_DESCRIPTION = (
    "Environment the service is running in (e.g., 'production', 'staging')."
)
SCHEMA_HEALTHZ_STATUS_DESCRIPTION = "Overall health status. True if all checks passed."
SCHEMA_HEALTHZ_LATENCY_MS_DESCRIPTION = (
    "Total time taken to run all health checks."
)
SCHEMA_HEALTHZ_DETAILS_DESCRIPTION = (
    "Breakdown of each individual check by name."
)



from pydantic import BaseModel, ConfigDict


class StrictModel(BaseModel):
    model_config = ConfigDict(
        extra='forbid',
        populate_by_name=True,
        validate_assignment=True,
        str_strip_whitespace=True,
        str_min_length=1
    )



from abc import ABC
from typing import Protocol, Self, runtime_checkable

from snakestack.pubsub.subscriber.batch import BatchSettingsModel
from snakestack.pubsub.subscriber.types import (
    AsyncQueueMessagePubSub,
    BufferPubSubType,
    DataPubSubType,
    MessagePubSubType,
    QueuesType,
)


@runtime_checkable
class SimpleProtocol(Protocol):
    async def process(self: Self, data: DataPubSubType, message: MessagePubSubType) -> None:
        ...

@runtime_checkable
class BatchProtocol(Protocol):

    _queues: QueuesType

    async def process(self: Self, data: DataPubSubType, message: MessagePubSubType) -> None: ...
    async def process_batch(self: Self, items: BufferPubSubType, batch_settings: BatchSettingsModel) -> None:
        ...

class SimpleProcessor(ABC):
    async def process(self: Self, data: DataPubSubType, message: MessagePubSubType) -> None:
        raise NotImplementedError("Missing implementation of 'process'. This method is required for Pub/Sub processing in custom subscribers.")

class BatchProcessor(SimpleProcessor):
    def __init__(self: Self) -> None:
        self._queues: QueuesType | None = None

    async def process_batch(self: Self, items: BufferPubSubType, batch_settings: BatchSettingsModel) -> None:
        raise NotImplementedError("Missing implementation of 'process_batch'. This method is required for batched Pub/Sub processing in custom subscribers.")

    def set_queues(self: Self, queues: QueuesType) -> None:
        self._queues = queues

    def get_queue(self: Self, context: str) -> AsyncQueueMessagePubSub:
        if not self._queues:
            raise AttributeError("Queue is not initialized")
        return self._queues[context]



from pydantic import Field

from snakestack import constants
from snakestack.model import StrictModel


class CheckModel(StrictModel):
    ok: bool = Field(..., description=constants.SCHEMA_CHECK_OK_DESCRIPTION)
    latency_ms: float = Field(
        ..., description=constants.SCHEMA_CHECK_LATENCY_MS_DESCRIPTION
    )
    error: str | None = Field(
        default=None, description=constants.SCHEMA_CHECK_ERROR_DESCRIPTION
    )


class HealthCheckModel(StrictModel):
    service_name: str = Field(
        ..., description=constants.SCHEMA_HEALTHZ_SERVICE_NAME_DESCRIPTION
    )
    version: str = Field(
        ..., description=constants.SCHEMA_HEALTHZ_VERSION_DESCRIPTION
    )
    host: str = Field(..., description=constants.SCHEMA_HEALTHZ_HOST_DESCRIPTION)
    uptime: str | None = Field(
        default=None, description=constants.SCHEMA_HEALTHZ_UPTIME_DESCRIPTION
    )
    timestamp: str = Field(
        ..., description=constants.SCHEMA_HEALTHZ_TIMESTAMP_DESCRIPTION
    )
    environment: str = Field(
        ..., description=constants.SCHEMA_HEALTHZ_ENVIRONMENT_DESCRIPTION
    )
    status: bool = Field(
        ..., description=constants.SCHEMA_HEALTHZ_STATUS_DESCRIPTION
    )
    latency_ms: float = Field(
        ..., description=constants.SCHEMA_HEALTHZ_LATENCY_MS_DESCRIPTION
    )
    details: dict[str, CheckModel] = Field(
        ..., description=constants.SCHEMA_HEALTHZ_DETAILS_DESCRIPTION
    )



import logging
import socket
from functools import lru_cache

try:
    from opentelemetry import trace
    from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import OTLPSpanExporter
    from opentelemetry.instrumentation.grpc import GrpcInstrumentorClient
    from opentelemetry.instrumentation.httpx import HTTPXClientInstrumentor
    from opentelemetry.instrumentation.logging import LoggingInstrumentor
    from opentelemetry.sdk.resources import SERVICE_NAME, Resource
    from opentelemetry.sdk.trace import TracerProvider
    from opentelemetry.sdk.trace.export import (
        BatchSpanProcessor,
        ConsoleSpanExporter,
        SimpleSpanProcessor,
    )
except ImportError:
    raise RuntimeError("Telemetry extra is not installed. Run `pip install snakestack[telemetry]`.")

from snakestack.config import settings

logger = logging.getLogger(__name__)


@lru_cache(maxsize=None)
def instrument_app(
    service_name: str,
    test_mode: bool = False,
    enable_httpx: bool = False,
    enable_grpc: bool = False,
    enable_mongodb: bool = False,
    logging_level: int = logging.INFO,
) -> None:
    """Instrumenta o app com OpenTelemetry padrão."""
    if settings.otel_sdk_disabled:
        logger.info("OpenTelemetry is disabled via OTEL_SDK_DISABLED.")
        return

    if settings.snakestack_otel_disabled:
        logger.info("OpenTelemetry is disabled via SNAKESTACK_OTEL_DISABLED.")
        return

    resource = Resource.create().merge(Resource(attributes={
        SERVICE_NAME: service_name,
        "service.instance.id": socket.gethostname(),
    }))

    provider = TracerProvider(resource=resource)
    exporter = ConsoleSpanExporter() if test_mode else OTLPSpanExporter(insecure=True)
    processor = SimpleSpanProcessor(exporter) if test_mode else BatchSpanProcessor(exporter)
    provider.add_span_processor(processor)
    trace.set_tracer_provider(provider)

    LoggingInstrumentor().instrument(set_logging_format=True, log_level=logging_level)

    if enable_httpx:
        logger.debug("Instrumenting HTTPX client...")
        HTTPXClientInstrumentor().instrument()

    if enable_grpc:
        logger.debug("Instrumenting gRPC client...")
        GrpcInstrumentorClient().instrument()

    if enable_mongodb:
        logger.debug("Instrumenting MongoDB client...")
        from snakestack.mongodb.patch import patch_motor_collection_methods
        patch_motor_collection_methods()



import functools
import inspect
from typing import Awaitable, Callable, ParamSpec, TypeVar, cast

try:
    from opentelemetry import trace
except ImportError:
    raise RuntimeError("Telemetry extra is not installed. Run `pip install snakestack[telemetry]`.")


_T = TypeVar("_T")
_P = ParamSpec("_P")


def get_current_span() -> trace.Span:
    return trace.get_current_span()


def get_tracer(name: str, version: str = "1.0.0") -> trace.Tracer:
    """Obtém o tracer padrão da aplicação."""
    return trace.get_tracer(name, version)



def instrumented_span(
    span_name: str
) -> Callable[[Callable[_P, _T] | Callable[_P, Awaitable[_T]]], Callable[_P, _T] | Callable[_P, Awaitable[_T]]]:
    """Decorator that creates a span with the given name for sync or async functions."""

    def decorator(
        func: Callable[_P, _T] | Callable[_P, Awaitable[_T]]
    ) -> Callable[_P, _T] | Callable[_P, Awaitable[_T]]:
        if inspect.iscoroutinefunction(func):
            @functools.wraps(func)
            async def async_wrapper(*args: _P.args, **kwargs: _P.kwargs) -> _T:
                with get_tracer(func.__module__).start_as_current_span(span_name):
                    return await cast(Callable[_P, Awaitable[_T]], func)(*args, **kwargs)
            return cast(Callable[_P, Awaitable[_T]], async_wrapper)

        else:
            @functools.wraps(func)
            def sync_wrapper(*args: _P.args, **kwargs: _P.kwargs) -> _T:
                with get_tracer(func.__module__).start_as_current_span(span_name):
                    return cast(Callable[_P, _T], func)(*args, **kwargs)
            return cast(Callable[_P, _T], sync_wrapper)

    return decorator


def set_span_attributes(**attrs: str | int | float | bool) -> None:
    span = get_current_span()
    for key, value in attrs.items():
        span.set_attribute(key, value)



import logging
from concurrent.futures import Future
from typing import Callable, Self

try:
    from google.cloud.pubsub_v1 import PublisherClient
except ImportError:
    raise RuntimeError("Pubsub extra is not installed. Run `pip install snakestack[pubsub]`.")

from snakestack.pubsub.exceptions import InvalidAttributesError, PublishError

logger = logging.getLogger(__name__)


class SnakeStackPublisher:

    def __init__(self: Self, client: "PublisherClient") -> None:
        self.client = client

    def publish(
        self: Self,
        project_id: str,
        topic_name: str,
        data: bytes,
        attributes: dict[str, str | bytes] | None = None,
        callback: Callable[[Future[None]], None] | None = None,
        timeout: int = 5
    ) -> None:
        topic_path = self.client.topic_path(project=project_id, topic=topic_name)
        attributes = attributes or {}

        for k, v in attributes.items():
            if not isinstance(v, (str, bytes)):
                raise InvalidAttributesError(key=k, value=v)

        future = self.client.publish(data=data, topic=topic_path, **attributes)

        if callback:
            future.add_done_callback(callback)
            logger.debug(f"Callback added to future for topic {topic_path}")
        else:
            try:
                future.result(timeout=timeout)
            except Exception as e:
                logger.exception("Error while publishing message.", exc_info=e)
                raise PublishError(topic=topic_path, original_exception=e)
            else:
                logger.debug(f"Successfully sent on topic {topic_path}")



import asyncio
import gzip
import json
import time
from typing import Any, Optional, Sequence, Type

try:
    from google.cloud.pubsub_v1 import SubscriberClient
    from google.cloud.pubsub_v1.subscriber.message import Message
except ImportError:
    raise RuntimeError("Pubsub extra is not installed. Run `pip install snakestack[pubsub]`.")

from pydantic import BaseModel, ValidationError

from .exceptions import AckableError


def parse_message(message: Message, schema: Type[BaseModel] | None = None) -> Any:
    raw_bytes = message.data
    encoding = message.attributes.get("encoding", "").lower()
    if encoding == "gzip":
        try:
            raw_data = gzip.decompress(raw_bytes).decode("utf-8")
        except Exception as e:
            raise AckableError(f"Error on decompress: {e}")
    else:
        raw_data = raw_bytes.decode("utf-8")

    try:
        parsed_data = json.loads(raw_data)
    except json.JSONDecodeError:
       return {"data": raw_data, "valid_json": False}

    if schema:
        try:
            parsed_data = schema(**parsed_data)
        except ValidationError as e:
            raise AckableError(f"Validation error: {e}")

    return {"valid_json": True, "data": parsed_data}

def get_message_info(message: Message) -> tuple[float, float]:
    publish_time = message.publish_time.timestamp()
    latency_ms = (time.time() - publish_time) * 1000
    message_kb = message.size / 1024
    return latency_ms, message_kb

async def ack_messages(
    messages: Sequence[Message],
    *,
    subscriber_client: Optional[SubscriberClient] = None,
    subscription_path: Optional[str] = None,
    batch_threshold: int = 100,
) -> None:
    if not messages:
        return

    if not subscriber_client or not subscription_path or len(messages) < batch_threshold:
        for msg in messages:
            msg.ack()
        return

    ack_ids = [msg.ack_id for msg in messages]
    loop = asyncio.get_running_loop()
    await loop.run_in_executor(
        None,
        subscriber_client.acknowledge,
        subscription_path,
        ack_ids,
    )


async def nack_messages(
    messages: Sequence[Message],
) -> None:
    if not messages:
        return

    for msg in messages:
        msg.nack()



from .async_client import (
    close_mongodb_connection,
    get_collection,
    get_database,
    open_mongodb_connection,
)
from .patch import patch_motor_collection_methods

__all__ = [
    "close_mongodb_connection",
    "open_mongodb_connection",
    "get_database",
    "get_collection",
    "patch_motor_collection_methods"
]



from typing import Self


class SnakeStackPubSubError(Exception):
    """Base exception for SnakeStack PubSub errors."""


class PublishError(SnakeStackPubSubError):
    """Raised when publishing a message fails."""

    def __init__(self: Self, topic: str, original_exception: Exception) -> None:
        super().__init__(f"Failed to publish message to topic '{topic}': {original_exception}")
        self.topic = topic
        self.original_exception = original_exception


class InvalidAttributesError(SnakeStackPubSubError):
    """Raised when attributes passed to publish are invalid."""

    def __init__(self: Self, key: str, value: object) -> None:
        super().__init__(f"Attribute '{key}' must be str or bytes, got {type(value).__name__}")
        self.key = key
        self.value = value



from typing import Self

try:
    from google.cloud.pubsub_v1.types import FlowControl
except ImportError:
    raise RuntimeError("Pubsub extra is not installed. Run `pip install snakestack[pubsub]`.")

from snakestack import constants


class PubSubFlowControlFactory:

    def __init__(self: Self) -> None:
        self._presets = {
            "low_latency": self.low_latency,
            "balanced": self.balanced,
            "high_throughput": self.high_throughput,
            "aggressive": self.aggressive,
        }

    def create(
        self: Self,
        max_messages: int,
        max_bytes: int,
        max_lease_duration: int
    ) -> FlowControl:
        return FlowControl(
            max_messages=max_messages,
            max_bytes=max_bytes,
            max_lease_duration=max_lease_duration
        )

    def low_latency(self: Self) -> FlowControl:
        return self.create(
            max_messages=1,
            max_bytes=10 * constants.KILOBYTE,
            max_lease_duration=10 * constants.SECONDS
        )

    def balanced(self: Self) -> FlowControl:
        return self.create(
            max_messages=10,
            max_bytes=100 * constants.KILOBYTE,
            max_lease_duration=1 * constants.MINUTES
        )

    def high_throughput(self: Self) -> FlowControl:
        return self.create(
            max_messages=1000,
            max_bytes=5 * constants.MEGABYTE,
            max_lease_duration=5 * constants.MINUTES
        )

    def aggressive(self: Self) -> FlowControl:
        return self.create(
            max_messages=5000,
            max_bytes=20 * constants.MEGABYTE,
            max_lease_duration=10 * constants.MINUTES
        )



from .config import setup_logging
from .contexts import get_request_id, reset_request_id, set_request_id
from .filters import ExcludeLoggerFilter, RequestIdFilter
from .formatters import FORMATTERS, JsonFormatter
from .handlers import HANDLERS

__all__ = [
    "FORMATTERS",
    "HANDLERS",
    "ExcludeLoggerFilter",
    "RequestIdFilter",
    "JsonFormatter",
    "setup_logging",
    "get_request_id",
    "set_request_id",
    "reset_request_id"
]



class BaseMessageError(Exception):
    """Erro genérico de processamento de mensagem."""


class AckableError(BaseMessageError):
    """Erro em que a mensagem deve ser ACKed mesmo com falha."""


class RetryableError(BaseMessageError):
    """Erro em que a mensagem deve ser NACKed para retry."""


class BlockRuleError(RetryableError):
    """Mensagem bloqueada por regra de negócio (deve ser NACKed)."""



import asyncio
import signal

from .client import SnakeStackSubscriber


def setup_signal_handlers(subscriber: SnakeStackSubscriber) -> None:
    loop = asyncio.get_running_loop()
    for sig in (signal.SIGINT, signal.SIGTERM):
        loop.add_signal_handler(sig, subscriber.shutdown)



from pydantic import BaseModel, ConfigDict, Field

from snakestack import constants


class ErrorModel(BaseModel):
    model_config = ConfigDict(
        extra='allow',
        str_strip_whitespace=True,
        str_min_length=1,
        frozen=True
    )
    detail: str = Field(..., description=constants.ERROR_DETAIL_SCHEMA_DESCRIPTION)



from functools import wraps
from typing import Any, Callable, Coroutine, TypeVar, cast

try:
    from opentelemetry.trace import SpanKind, StatusCode
except ImportError:
    raise RuntimeError("Telemetry extra is not installed. Run `pip install snakestack[telemetry]`.")

from snakestack.telemetry import get_tracer

F = TypeVar("F", bound=Callable[..., Coroutine[Any, Any, Any]])
tracer = get_tracer("snakestack.pubsub.processor")

def traced_processor(operation_name: str) -> Callable[[F], F]:
    def decorator(func: F) -> F:
        @wraps(func)
        async def wrapper(*args: Any, **kwargs: Any) -> Any:
            with tracer.start_as_current_span(operation_name, kind=SpanKind.INTERNAL) as span:
                try:
                    if "message" in kwargs:
                        span.set_attribute("message.id", kwargs["message"].message_id)
                    if "items" in kwargs:
                        span.set_attribute("batch.size", len(kwargs["items"]))
                    return await func(*args, **kwargs)
                except Exception as e:
                    span.set_status(StatusCode.ERROR)
                    span.record_exception(e)
                    raise

        return cast(F, wrapper)
    return decorator



import logging
import logging.config

from snakestack.config import settings
from snakestack.logging.formatters import FORMATTERS
from snakestack.logging.handlers import HANDLERS

DEFAULT_LOGGING_CONFIG = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": FORMATTERS,
    "handlers": HANDLERS,
    "root": {
        "level": settings.snakestack_log_level,
        "handlers": ["console"]
    }
}


def setup_logging() -> None:
    logging.config.dictConfig(DEFAULT_LOGGING_CONFIG)



from http import HTTPStatus
from typing import Any, Type

from pydantic import BaseModel

from .error import ErrorModel

ResponseDoc = dict[int | str, dict[str, Any]]

def _create_http_status_code(status_code: int) -> HTTPStatus:
    status = HTTPStatus(status_code)
    if not any([
        status.is_success,
        status.is_redirection,
        status.is_informational,
        status.is_server_error,
        status.is_client_error]
    ):
        raise ValueError(f"Unsupported status code: {status_code}")
    return status


def make_standard_response(
    status_code: int,
    *,
    model: Type[BaseModel],
    example: dict[str, Any],
    content_type: str = "application/json",
    override_description: str | None = None,
) -> ResponseDoc:
    status = _create_http_status_code(status_code=status_code)

    result = {
        "model": model,
        "description": override_description or status.description,
        "content": {
            content_type: {
                "example": example
            }
        },
    }

    if status_code == 204:
        result.pop("model")
        result.pop("content")

    return {
        status_code: result
    }


def make_standard_error_response(
    status_code: int,
    *,
    model: Type[BaseModel] | None = None,
    content_type: str = "application/json",
    example: dict[str, Any] | None = None,
    override_description: str | None = None,
    override_example: str | None = None
) -> ResponseDoc:
    """
    Gera uma resposta OpenAPI com modelo e exemplo opcional.
    """
    model = model or ErrorModel
    desc_error = override_example or "Error"
    example = example or ErrorModel(detail=desc_error).model_dump()
    return make_standard_response(
        status_code=status_code,
        model=model or ErrorModel,
        override_description=override_description,
        example=example,
        content_type=content_type
    )


def make_standard_success_response(
    status_code: int,
    model: Type[BaseModel],
    *,
    description: str | None = None,
    example: dict[str, Any] | None = None,
    content_type: str = "application/json"
) -> ResponseDoc:
    """
    Gera uma resposta OpenAPI com modelo e exemplo opcional.
    """
    example = example or {"foo": "bar"}
    return make_standard_response(
        status_code=status_code,
        model=model,
        override_description=description,
        example=example,
        content_type=content_type
    )



from functools import wraps
from typing import Any, Callable, Coroutine, TypeVar, cast

try:
    from opentelemetry.trace import SpanKind, get_tracer
    from opentelemetry.trace.status import Status, StatusCode
except ImportError:
    raise RuntimeError("Telemetry extra is not installed. Run `pip install snakestack[telemetry]`.")

F = TypeVar("F", bound=Callable[..., Coroutine[Any, Any, Any]])
tracer = get_tracer(__name__)

def traced_motor_method(operation_name: str) -> Callable[[F], F]:
    """
    Wrapper para criar um span em volta da operação do Motor.
    Pode ser aplicado dinamicamente nos métodos como find, insert_one etc.
    """

    def decorator(func: F) -> F:
        @wraps(func)
        async def wrapper(self: Any, *args: Any, **kwargs: Any) -> Any:
            collection_name = getattr(self, 'name', 'unknown_collection')

            with tracer.start_as_current_span(
                f"mongo.{operation_name}",
                kind=SpanKind.CLIENT,
            ) as span:
                span.set_attribute("db.system", "mongodb")
                db_name = getattr(getattr(self, "database", None), "name", "unknown_db")
                span.set_attribute("db.name", db_name)
                span.set_attribute("db.mongodb.collection", collection_name)
                span.set_attribute("db.operation", operation_name)

                if args:
                    span.set_attribute("db.mongo.args_count", len(args))
                if kwargs:
                    span.set_attribute("db.mongo.kwargs_keys", len(kwargs.keys()))

                try:
                    result = await func(self, *args, **kwargs)
                    span.set_status(Status(StatusCode.OK))
                    return result
                except Exception as exc:
                    span.set_status(Status(StatusCode.ERROR, str(exc)))
                    span.record_exception(exc)
                    raise

        setattr(wrapper, "_is_traced_motor_method", True)
        return cast(F, wrapper)

    return decorator



from typing import Self


class SnakeHTTPError(Exception):
    """Base exception for SnakeStack HTTP errors."""


class RequestHTTPError(SnakeHTTPError):
    """Raised when requesting fails."""

    def __init__(self: Self, api: str, original_exception: Exception) -> None:
        super().__init__(f"Failed to request to api '{api}': {original_exception}")
        self.api = api
        self.original_exception = original_exception



import functools
import logging
from typing import Any, Awaitable, Callable, ParamSpec, TypeVar

try:
    from redis.exceptions import RedisError
except ImportError:
    raise RuntimeError("Redis extra is not installed. Run `pip install snakestack[redis]`.")

logger = logging.getLogger(__name__)

T = TypeVar("T")
P = ParamSpec("P")


def deco_cache(
    default: Any = None
) -> Callable[[Callable[P, Awaitable[T]]], Callable[P, Awaitable[T | Any]]]:
    def decorator(fn: Callable[P, Awaitable[T]]) -> Callable[P, Awaitable[T | Any]]:
        @functools.wraps(fn)
        async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T | Any:
            try:
                return await fn(*args, **kwargs)
            except RedisError as error:
                logger.exception("Redis operation failed.", exc_info=error)
                return default
        return wrapper
    return decorator



import asyncio
import logging
import time
from typing import Awaitable, Callable, Self, TypeAlias

try:
    from google.cloud.pubsub_v1 import SubscriberClient
    from google.cloud.pubsub_v1.subscriber.message import Message
except ImportError:
    raise RuntimeError("Pubsub extra is not installed. Run `pip install snakestack[pubsub]`.")

from snakestack.model import LenientModel
from snakestack.pubsub.subscriber.types import AsyncQueueMessagePubSub, DataPubSubType

logger = logging.getLogger(__name__)


class BatchSettingsModel(LenientModel):
    size: int
    timeout: float
    name: str
    subscriber_client: SubscriberClient | None = None
    subscription_path: str | None = None


class BatchingWorker:
    def __init__(
        self: Self,
        *,
        callback: Callable[[list[tuple[Message, DataPubSubType]], BatchSettingsModel], Awaitable[None]],
        batch_settings: BatchSettingsModel
    ) -> None:
        self.buffer: list[tuple[Message, DataPubSubType]] = []
        self.last_flush = time.monotonic()
        self.callback = callback
        self.batch_settings = batch_settings

    async def run(self: Self, queue: AsyncQueueMessagePubSub) -> None:
        logger.debug(f"Starting batch [{self.batch_settings.name}]")
        while True:
            try:
                timeout = max(self.batch_settings.timeout - (time.monotonic() - self.last_flush), 0.1)
                message, parsed = await asyncio.wait_for(queue.get(), timeout=timeout)
                self.buffer.append((message, parsed))
            except asyncio.TimeoutError:
                pass

            if len(self.buffer) >= self.batch_settings.size or (time.monotonic() - self.last_flush) >= self.batch_settings.timeout:
                await self._flush()

    async def _flush(self: Self) -> None:
        if not self.buffer:
            return

        logger.info(f"[{self.batch_settings.name}] Executing batch with {len(self.buffer)} messages")
        try:
            await self.callback(self.buffer, self.batch_settings)
        except Exception as e:
            logger.exception("Failed to process batch", exc_info=e)
        finally:
            self.buffer.clear()
            self.last_flush = time.monotonic()

BatchesType: TypeAlias = dict[str, BatchingWorker]



import asyncio
from typing import Any, Type, TypeAlias

try:
    from google.cloud.pubsub_v1.subscriber.message import Message
except ImportError:
    raise RuntimeError("Pubsub extra is not installed. Run `pip install snakestack[pubsub]`.")

from pydantic import BaseModel

AttributesPubSubType: TypeAlias = dict[str, str]
MessagePubSubType: TypeAlias = Message
DataPubSubType: TypeAlias = dict[str, Any]
BufferPubSubType: TypeAlias = list[tuple[Message, DataPubSubType]]
AsyncQueueMessagePubSub: TypeAlias = asyncio.Queue[MessagePubSubType]
QueuesType: TypeAlias = dict[str, AsyncQueueMessagePubSub]
SchemaType: TypeAlias = Type[BaseModel]



import json
from datetime import UTC, datetime
from logging import Formatter, LogRecord
from typing import Any, Self

from snakestack.logging.encoders import safe_jsonable_encoder

FORMATTERS = {
    "default": {
        "format": "%(asctime)s [%(levelname)s] %(name)s: %(message)s"
    },
    "with_request_id": {
        "format": (
            "%(asctime)s [%(levelname)s] [req_id=%(request_id)s] "
            "%(name)s: %(message)s"
        )
    }
}


class JsonFormatter(Formatter):
    def format(
        self: Self,
        record: LogRecord,
        *args: Any,
        **kwargs: Any
    ) -> str:
        Formatter.format(self, record)
        return json.dumps(safe_jsonable_encoder(self._create_log(record)))

    def _create_log(self: Self, record: LogRecord) -> dict[str, Any]:
        message: dict[str, Any] = {
            "time": datetime.fromtimestamp(record.created, tz=UTC).isoformat(),
            "level": record.levelname,
            "pid": record.process,
            "name": f"{record.name}:{record.lineno}",
            "msg": record.getMessage(),
            "request": {}
        }

        if record.exc_info:
            message["exc_info"] = str(record.exc_info)

        if hasattr(record, "request_id") and record.request_id:
            message["request"].update({"id": record.request_id})

        if hasattr(record, "otelTraceID") and record.otelTraceID:
            message["request"].update({"trace_id": record.otelTraceID})

        if hasattr(record, "otelSpanID") and record.otelSpanID:
            message["request"].update({"span_id": record.otelSpanID})

        return message



from typing import Self

try:
    from google.api_core.retry import Retry
    from google.cloud.pubsub_v1 import PublisherClient
    from google.cloud.pubsub_v1.types import (
        BatchSettings,
        LimitExceededBehavior,
        PublisherOptions,
        PublishFlowControl,
    )
except ImportError:
    raise RuntimeError("Pubsub extra is not installed. Run `pip install snakestack[pubsub]`.")

from snakestack import constants


class PubSubPublisherFactory:

    def __init__(self: Self, enable_open_telemetry: bool = False) -> None:
        self.enable_open_telemetry = enable_open_telemetry

    def create(
        self: Self,
        batch_settings: "BatchSettings",
        publisher_options: "PublisherOptions"
    ) -> PublisherClient:
        return PublisherClient(
            batch_settings=batch_settings,
            publisher_options=publisher_options,
        )

    def low_latency(self: Self) -> PublisherClient:
        """
        Cria um publisher otimizado para baixa latência.

        Ideal para eventos críticos que exigem envio quase imediato,
        como auditoria, billing ou notificações sensíveis ao tempo.

        Configurações:
        - Envia até 10 mensagens por batch com no máximo 10ms de espera.
        - Tamanho máximo do batch: 64KB.
        - Permite até 100 mensagens ou 1MB de mensagens pendentes no cliente.
        - Se exceder o fluxo, uma exceção será lançada (comportamento estrito).
        - Retry com timeout curto (5s) e backoff rápido.
        """
        batch_settings = BatchSettings(
            max_messages=10,
            max_latency=10 * constants.MILLISECONDS,
            max_bytes=64 * constants.KILOBYTE
        )

        publisher_options = PublisherOptions(
            flow_control=PublishFlowControl(
                message_limit=100,
                byte_limit=constants.MEGABYTE,
                limit_exceeded_behavior=LimitExceededBehavior.ERROR
            ),
            retry=Retry(
                initial=0.05,
                multiplier=1.2,
                maximum=5 * constants.SECONDS,
                timeout=5 * constants.SECONDS
            ),
            enable_open_telemetry_tracing=self.enable_open_telemetry
        )

        return self.create(
            batch_settings=batch_settings,
            publisher_options=publisher_options
        )

    def high_throughput(self: Self) -> PublisherClient:
        """
        Cria um publisher otimizado para throughput máximo.

        Ideal para processamento em lote, observabilidade ou
        sistemas com grandes volumes de mensagens e menor sensibilidade à latência.

        Configurações:
        - Até 1000 mensagens por batch, 10MB, ou 100ms de espera.
        - Buffer de até 5000 mensagens ou 50MB.
        - Bloqueia publicações quando limites são excedidos.
        - Retry com deadline longo (90s) e backoff agressivo.
        """
        batch_settings = BatchSettings(
            max_messages=1000,
            max_latency=100 * constants.MILLISECONDS,
            max_bytes=10 * constants.MEGABYTE
        )

        publisher_options = PublisherOptions(
            flow_control=PublishFlowControl(
                message_limit=5000,
                byte_limit=50 * constants.MEGABYTE,
                limit_exceeded_behavior=LimitExceededBehavior.BLOCK
            ),
            retry=Retry(
                initial=0.2,
                multiplier=1.5,
                maximum=60 * constants.SECONDS,
                timeout=90 * constants.SECONDS
            ),
            enable_open_telemetry_tracing=self.enable_open_telemetry
        )

        return self.create(
            batch_settings=batch_settings,
            publisher_options=publisher_options
        )

    def balanced(self: Self) -> PublisherClient:
        """
        Cria um publisher balanceado entre latência e throughput.

        Ideal para cenários padrão de produção com volume moderado
        e necessidade de desempenho confiável.

        Configurações:
        - Até 500 mensagens por batch, 1MB, ou 50ms de espera.
        - Até 1000 mensagens ou 10MB no buffer do cliente.
        - Bloqueia quando excede o limite de fluxo (comportamento seguro).
        - Retry configurado com deadline de 60 segundos e backoff exponencial.
        """
        batch_settings = BatchSettings(
            max_messages=500,
            max_latency=50 * constants.MILLISECONDS,
            max_bytes=constants.MEGABYTE
        )

        publisher_options = PublisherOptions(
            flow_control=PublishFlowControl(
                message_limit=1000,
                byte_limit=10 * constants.MEGABYTE,
                limit_exceeded_behavior=LimitExceededBehavior.BLOCK
            ),
            retry=Retry(
                initial=0.1,
                multiplier=1.3,
                maximum=30 * constants.SECONDS,
                timeout=60 * constants.SECONDS
            ),
            enable_open_telemetry_tracing=self.enable_open_telemetry
        )

        return self.create(
            batch_settings=batch_settings,
            publisher_options=publisher_options
        )

    def aggressive(self: Self) -> PublisherClient:
        """
        Cria um publisher agressivo, ideal para testes de carga e ambientes controlados.

        Usa limites extremamente altos e não bloqueia nem lança exceções ao ultrapassar
        o fluxo, favorecendo performance e volume.

        Configurações:
        - Até 2000 mensagens por batch, 10MB, ou 500ms de espera.
        - Buffer com até 10000 mensagens ou 100MB.
        - Ignora excessos de fluxo (comportamento arriscado).
        - Retry rápido com timeout moderado (20s).
        """
        batch_settings = BatchSettings(
            max_messages=2000,
            max_latency=500 * constants.MILLISECONDS,
            max_bytes=10 * constants.MEGABYTE
        )

        publisher_options = PublisherOptions(
            flow_control=PublishFlowControl(
                message_limit=10000,
                byte_limit=100 * constants.MEGABYTE,
                limit_exceeded_behavior=LimitExceededBehavior.IGNORE
            ),
            retry=Retry(
                initial=0.1,
                multiplier=2.0,
                maximum=10 * constants.SECONDS,
                timeout=20 * constants.SECONDS
            ),
            enable_open_telemetry_tracing=self.enable_open_telemetry
        )

        return self.create(
            batch_settings=batch_settings,
            publisher_options=publisher_options
        )



from datetime import datetime
from decimal import Decimal
from enum import Enum
from pathlib import Path
from typing import Any
from uuid import UUID


def safe_jsonable_encoder(obj: Any) -> Any:
    if isinstance(obj, (str, int, float, bool, type(None))):
        return obj

    if isinstance(obj, (datetime, UUID, Enum, Decimal, Path)):
        return str(obj)

    if isinstance(obj, dict):
        return {safe_jsonable_encoder(k): safe_jsonable_encoder(v) for k, v in obj.items()}

    if isinstance(obj, (list, tuple, set, frozenset)):
        return [safe_jsonable_encoder(item) for item in obj]

    if hasattr(obj, "dict"):
        return safe_jsonable_encoder(obj.dict())

    if hasattr(obj, "model_dump"):
        return safe_jsonable_encoder(obj.model_dump())

    if hasattr(obj, "__dict__"):
        return safe_jsonable_encoder(vars(obj))

    return str(obj)



import logging
from typing import Any

try:
    from motor.motor_asyncio import (
        AsyncIOMotorClient,
        AsyncIOMotorCollection,
        AsyncIOMotorDatabase,
    )
except ImportError:
    raise RuntimeError("Mongodb extra is not installed. Run `pip install snakestack[mongodb]`.")

from snakestack.config import settings


class DB:
    client: AsyncIOMotorClient[Any] | None = None

db = DB()

logger = logging.getLogger(__name__)


async def open_mongodb_connection() -> None:
    logger.debug("Connecting to mongodb...")
    if not settings.snakestack_mongodb_url.startswith("mongodb"):
        raise ValueError("Invalid MongoDB URL configured.")
    db.client = AsyncIOMotorClient(settings.snakestack_mongodb_url)
    logger.debug("Connecting to mongodb successful.")


async def close_mongodb_connection() -> None:
    logger.debug("Closing connection with mongodb...")
    if db.client:
        db.client.close()
    logger.debug("Connection with mongodb is closed.")


def get_collection(collection: str) -> AsyncIOMotorCollection[Any]:
    if not db.client:
        raise RuntimeError("Connection with mongodb is not starting.")
    return db.client[settings.snakestack_mongodb_dbname][collection]


def get_database() -> AsyncIOMotorDatabase[Any]:
    if not db.client:
        raise RuntimeError("Connection with mongodb is not starting.")
    return db.client[settings.snakestack_mongodb_dbname]



from .base import StrictModel
from .doc import (
    ResponseDoc,
    make_standard_error_response,
    make_standard_success_response,
)
from .error import ErrorModel
from .lenient import LenientModel
from .orm import ORMModel
from .pagination import PaginatedModel

__all__ = [
    "StrictModel",
    "LenientModel",
    "PaginatedModel",
    "ORMModel",
    "ErrorModel",
    "make_standard_error_response",
    "make_standard_success_response",
    "ResponseDoc"
]



import json
from typing import Any, Self, cast

try:
    from redis.asyncio import Redis
except ImportError:
    raise RuntimeError("Redis extra is not installed. Run `pip install snakestack[redis]`.")
from snakestack.cache.utils import deco_cache


class AsyncRedisService:
    """
    Serviço assíncrono para operações básicas com Redis,
    com suporte a prefixo de chave, TTL padrão e serialização JSON.
    """

    def __init__(
        self: Self,
        client: Redis,
        default_ttl: int | None = None,
        prefix: str = "",
    ) -> None:
        """
        Inicializa o serviço Redis.

        Args:
            client: Instância de redis.asyncio.Redis já conectada.
            default_ttl: Tempo padrão de expiração em segundos (opcional).
            prefix: Prefixo opcional para organizar as chaves.
        """
        self._client = client
        self._default_ttl = default_ttl
        self._prefix = prefix.strip(":")

    def _format_key(self: Self, key: str) -> str:
        """Adiciona o prefixo à chave, se configurado."""
        return f"{self._prefix}:{key}" if self._prefix else key

    @deco_cache(default=False)
    async def set(self: Self, key: str, value: Any, ex: int | None = None) -> bool:
        """
        Armazena um valor serializado em Redis com TTL opcional.

        Args:
           key: Nome da chave.
           value: Valor a ser serializado.
           ex: Expiração em segundos (override do TTL padrão).

        Returns:
           True se armazenado com sucesso, False em caso de erro.
        """
        serialized = json.dumps(value)
        return bool(await self._client.set(
            name=self._format_key(key),
            value=serialized,
            ex=ex or self._default_ttl,
        ))

    @deco_cache(default=None)
    async def get(self: Self, key: str) -> Any | None:
        """
        Recupera e desserializa um valor armazenado em Redis.

        Args:
            key: Nome da chave.

        Returns:
            Valor desserializado ou None.
        """
        raw = await self._client.get(self._format_key(key))
        if raw is None:
            return None
        try:
            return json.loads(raw)
        except json.JSONDecodeError:
            return raw

    @deco_cache(default=0)
    async def delete(self: Self, key: str) -> int:
        """Remove uma chave do Redis. Retorna 1 se removida, 0 se não existe."""
        result = await self._client.delete(self._format_key(key))
        return cast(int, result)

    @deco_cache(default=False)
    async def exists(self: Self, key: str) -> bool:
        """Verifica se a chave existe em Redis."""
        return bool(await self._client.exists(self._format_key(key)))

    @deco_cache(default=False)
    async def expire(self: Self, key: str, ttl: int) -> bool:
        """Define um tempo de expiração (em segundos) para uma chave."""
        return bool(await self._client.expire(self._format_key(key), ttl))

    @deco_cache(default=0)
    async def incr(self: Self, key: str, amount: int = 1) -> int:
        """Incrementa um valor numérico armazenado na chave."""
        result = await self._client.incr(self._format_key(key), amount)
        return cast(int, result)

    @deco_cache(default=0)
    async def decr(self: Self, key: str, amount: int = 1) -> int:
        """Decrementa um valor numérico armazenado na chave."""
        result = await self._client.decr(self._format_key(key), amount)
        return cast(int, result)

    @deco_cache(default=0)
    async def ttl(self: Self, key: str) -> int:
        """Obtém o tempo de expiração restante da chave, em segundos."""
        result = await self._client.ttl(self._format_key(key))
        return cast(int, result)



import inspect
import logging
import socket
import time
from datetime import UTC, datetime
from typing import Any, Awaitable, Callable, Self

from .model import CheckModel, HealthCheckModel

logger = logging.getLogger(__name__)

class SnakeHealthCheck:

    UPTIME_DIR = "/proc/uptime"

    def __init__(
        self: Self,
        service_name: str,
        service_version: str,
        service_environment: str,
    ) -> None:
        self._checks: dict[str, Callable[[], bool | Awaitable[bool]]] = {}
        self._service_name = service_name
        self._service_version = service_version
        self._service_environment = service_environment

    async def is_healthy(self: Self) -> tuple[dict[str, Any], bool]:
        result = {}
        timestamp = datetime.now(UTC).isoformat()
        uptime = self._uptime()

        checks: list[bool] = list()
        for name, check_fn in self._checks.items():
            start_check = time.perf_counter()
            ok, error = await self._safe_call(fn=check_fn)
            latency = round((time.perf_counter() - start_check) * 1000, 2)

            result[name] = CheckModel(
                ok=ok,
                latency_ms=latency,
                error=error if error else None
            ).model_dump()
            checks.append(ok)

        all_ok = all(checks)

        total_latency = sum(dep["latency_ms"] for dep in result.values())
        data = {
            "service_name": self._service_name,
            "version": self._service_version,
            "environment": self._service_environment,
            "host": socket.gethostname(),
            "timestamp": timestamp,
            "uptime": uptime,
            "status": all_ok,
            "latency_ms": total_latency,
            "details": result,
        }
        model = HealthCheckModel(**data).model_dump(exclude_none=True)
        return model, all_ok

    async def _safe_call(
        self: Self, fn: Callable[[], bool | Awaitable[bool]]
    ) -> tuple[bool, Any]:
        try:
            if inspect.iscoroutinefunction(fn):
                result = await fn()
            else:
                result = fn()
            return bool(result), None
        except Exception as error:
            logger.exception("Error on health check function", exc_info=error)
            return False, str(error)

    def add_check(self: Self, name: str, func: Callable[[], bool | Awaitable[bool]]) -> None:
        if name not in self._checks:
            self._checks[name] = func

    def _uptime(self: Self) -> str | None:
        try:
            with open(self.UPTIME_DIR) as f:
                uptime_seconds = float(f.readline().split()[0])
                hours, remainder = divmod(int(uptime_seconds), 3600)
                minutes, seconds = divmod(remainder, 60)
                return f"{hours}h {minutes}m {seconds}s"
        except Exception:
            return None



import logging
from typing import Self

from snakestack.logging import get_request_id


class RequestIdFilter(logging.Filter):
    def filter(self: Self, record: logging.LogRecord) -> bool:
        record.request_id = get_request_id()
        return True

class ExcludeLoggerFilter(logging.Filter):
    def __init__(self: Self, excluded_name: list[str]) -> None:
        super().__init__()
        self.excluded = set(excluded_name)

    def filter(self: Self, record: logging.LogRecord) -> bool:
        return record.name not in self.excluded



from pydantic import BaseModel, ConfigDict


class LenientModel(BaseModel):
    model_config = ConfigDict(
        extra='allow',
        populate_by_name=True,
        arbitrary_types_allowed=True,
    )



from functools import lru_cache

from pydantic import Field
from pydantic_settings import BaseSettings, SettingsConfigDict

from snakestack import version


class SnakeStackSettings(BaseSettings):
    """
    Global configuration for the SnakeStack library.

    OpenTelemetry configuration:
    - `otel_sdk_disabled`: Disables the entire OpenTelemetry SDK. Equivalent to setting `OTEL_SDK_DISABLED=true`.
      Use this when you want to fully disable all tracing, metrics, and logs generated by OpenTelemetry.

    - `snakestack_otel_disabled`: Disables only the OpenTelemetry instrumentation provided by SnakeStack.
      This is useful when you already use OpenTelemetry via an external tool like `opentelemetry-bootstrap`,
      or a custom distribution (distro), and want to prevent duplicate instrumentation.
      Equivalent to setting `SNAKESTACK_OTEL_DISABLED=true`.

    These options can be combined with standard OpenTelemetry environment variables
    like `OTEL_RESOURCE_ATTRIBUTES` and `OTEL_EXPORTER_OTLP_HEADERS`.
    """
    model_config = SettingsConfigDict(env_file=".env", env_file_encoding="utf-8")

    snakestack_log_level: str = Field(
        default="INFO",
        description="Logging level for the application (e.g., DEBUG, INFO, WARNING, ERROR)."
    )

    snakestack_version: str = Field(
        default=version.__version__,
        description=""
    )

    snakestack_mongodb_url: str = Field(
        default="mongodb://localhost:27017",
        description=""
    )

    snakestack_mongodb_dbname: str = Field(
        default="snakestack",
        description=""
    )

    snakestack_otel_disabled: bool = Field(
        default=False,
        description=(
            "Disables only the OpenTelemetry instrumentation provided by the SnakeStack library. "
            "This allows external OpenTelemetry configurations (e.g., opentelemetry-bootstrap or custom distro) "
            "to take full control over instrumentation without duplication. "
            "Equivalent to setting the environment variable SNAKESTACK_OTEL_DISABLED=true."
        )
    )

    pubsub_project_id: str = Field(
        default="snakestack-project",
        description=""
    )

    otel_sdk_disabled: bool = Field(
        default=False,
        description=(
            "Disables the entire OpenTelemetry SDK, including all automatic and manual instrumentation. "
            "Equivalent to setting the environment variable OTEL_SDK_DISABLED=true. "
            "When enabled, no traces, metrics, or logs will be generated or exported by the OpenTelemetry SDK."
        )
    )


@lru_cache
def get_settings() -> SnakeStackSettings:
    return SnakeStackSettings()


settings = get_settings()
